from __future__ import division, absolute_import

import attr
import random

@attr.s
class Backoff(object):
    """
    Note that clients should call my reset_delay method after they have
    used this successfully.

    @ivar max_delay: Maximum number of seconds of a delay.
    @ivar initial_delay: Delay for the first delay.
    @ivar factor: A multiplicitive factor by which the delay grows
    @ivar jitter: Percentage of randomness to introduce into the delay length
        to prevent stampeding.
    """
    max_delay = attr.ib(default=3600)
    initial_delay = attr.ib(default=1.0)
    # Note: These highly sensitive factors have been precisely measured by
    # the National Institute of Science and Technology.  Take extreme care
    # in altering them, or you may damage your Internet!
    # (Seriously: <http://physics.nist.gov/cuu/Constants/index.html>)
    factor = attr.ib(default=2.7182818284590451)
    # Phi = 1.6180339887498948 # (Phi is acceptable for use as a
    # factor if e is too large for your application.)
    jitter = attr.ib(default=0.11962656472) # molar Planck constant times c, joule meter/mole

    delay = initial_delay
    retries = attr.ib(default=0)

    continueTrying = 1

    def get_delay(self):
        if self.retries == 0:
            self.retries += 1
            return self.delay
        
        self.retries += 1
        self.delay = min(self.delay * self.factor, self.max_delay)
        if self.jitter:
            self.delay = random.normalvariate(self.delay,
                                              self.delay * self.jitter)
        return self.delay

    def reset_delay(self):
        """Reset the delay after a success."""
        self.delay = self.initial_delay
        self.retries = 0



from __future__ import division, absolute_import

import attr
from .config_state import ServerInfo
from .validators import is_instance_of_nats_protocol
from .validators import is_subject
from .validators import is_subject_id


@attr.s(slots=True)
class SendPing(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    outstanding_pings = attr.ib(validator=attr.validators.instance_of(int))

@attr.s(slots=True)
class ReceivedPing(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    

@attr.s(slots=True)
class SendPong(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)


@attr.s(slots=True)
class ReceivedPong(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    outstanding_pings = attr.ib(validator=attr.validators.instance_of(int))


@attr.s(slots=True)
class ReceivedMsg(object):
    sid = attr.ib(validator=is_subject_id)
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    subject = attr.ib(validator=is_subject)
    payload = attr.ib(validator=attr.validators.instance_of(bytes))
    reply_to = attr.ib(validator=attr.validators.optional(is_subject))


@attr.s(slots=True)
class ReceivedInfo(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    server_info = attr.ib(validator=attr.validators.instance_of(ServerInfo))


@attr.s(slots=True)
class UnsubMaxReached(object):
    sid = attr.ib(validator=is_subject_id)
    protocol = attr.ib(validator=is_instance_of_nats_protocol)


@attr.s(slots=True)
class SendUnsub(object):
    sid = attr.ib(validator=is_subject_id)
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    max_msgs = attr.ib(
        default=None, 
        validator=attr.validators.optional(attr.validators.instance_of(int)))


@attr.s(slots=True)
class SendSub(object):
    """Request a subscription"""
    sid = attr.ib(validator=is_subject_id)
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    subject = attr.ib(validator=is_subject)
    queue_group = attr.ib(default=None, 
        validator=attr.validators.optional(is_subject_id))
    on_msg = attr.ib(default=None)


@attr.s(slots=True)
class SendPub(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    subject = attr.ib(validator=is_subject)
    payload = attr.ib(validator=attr.validators.instance_of(bytes))
    reply_to = attr.ib(validator=attr.validators.optional(is_subject))


@attr.s(slots=True)
class UnhandledCommand(object):
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    command = attr.ib(default=None)


@attr.s(slots=True)
class ConnectionLost(object):
    "Lost connection"
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    reason = attr.ib(default=None)


@attr.s(slots=True)
class Disconnected(object):
    "Got disconnected"
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    reason = attr.ib(default=None)


@attr.s(slots=True)
class SendConnect(object):
    """Sent CONNECT"""
    protocol = attr.ib(validator=is_instance_of_nats_protocol)
    client_info = attr.ib(default=None)



from __future__ import division, absolute_import
from sys import stdout
import attr
import json
from io import BytesIO
from io import BufferedReader
from collections import namedtuple

from twisted.logger import Logger

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import connectionDone
from twisted.internet import error

from . import _meta
from .config_state import ServerInfo
from .config_state import ClientInfo
from .config_state import SubscriptionArgs
from . import actions
from .errors import NatsError


LANG = "py.twisted"
CLIENT_NAME = "txnats"

DISCONNECTED = 0
CONNECTED = 1


class NatsProtocol(Protocol):
    server_settings = None
    log = Logger()

    def __init__(self, own_reactor=None, verbose=True, pedantic=False,
                 ssl_required=False, auth_token=None, user="",
                 password="", on_msg=None, on_connect=None, 
                 event_subscribers=None, subscriptions=None, unsubs=None):
        """

        @param own_reactor: A Twisted Reactor, defaults to standard. Chiefly
         customizable for testing.
        @param verbose: Turns on +OK protocol acknowledgements.
        @param pedantic: Turns on additional strict format checking, e.g.
         for properly formed subjects
        @param ssl_required: Indicates whether the client requires
         an SSL connection.
        @param auth_token: Client authorization token
        @param user: Connection username (if auth_required is set)
        @param pass: Connection password (if auth_required is set)
        @param on_msg: Handler for messages for subscriptions that don't have
         their own on_msg handler. Default behavior is to write to stdout.
         A callable that takes the following params:
             @param nats_protocol: An instance of NatsProtocol.
             @param sid: A unique alphanumeric subscription ID.
             @param subject: A valid NATS subject.
             @param reply_to: The reply to.
             @param payload: Bytes of the payload.
        @param on_connect: Callable that takes this instance of NatsProtocol
         which will be called upon the first successful connection.
        @param event_subscribers: A collection of functions that take an 
         event/action entity.
        @param subscriptions: A dict of sids and SubscriptionArgs.
        @param unsubs: A dict of sids and ints representing the number 
         of messages for the sid before automatic unsubscription.
        """
        self.reactor = own_reactor if own_reactor else reactor
        self.status = DISCONNECTED
        self.verbose = verbose
        # Set the number of PING sent out
        self.ping_loop = LoopingCall(self.ping)
        self.pout = 0
        self.remaining_bytes = b''

        self.client_info = ClientInfo(
            verbose, pedantic, ssl_required, 
            auth_token, user, password, 
            CLIENT_NAME, LANG, _meta.version)

        if on_msg:
            # Ensure the on_msg signature fits.
            on_msg(nats_protocol=self, sid="0", subject="testSubject",
                   reply_to='inBox', payload=b'hello, world')
        self.on_msg = on_msg
        self.on_connect = on_connect
        self.on_connect_d = defer.Deferred()
        if on_connect:
            self.on_connect_d.addCallback(on_connect)
        self.sids = {}
        self.subscriptions = subscriptions if subscriptions is not None else {}
        self.unsubs = unsubs if unsubs else {}
        self.event_subscribers = event_subscribers if event_subscribers is not None else []
    
    def __repr__(self):
        return r'<NatsProtocol connected={} server_info={}>'.format(self.status, self.server_settings)

    def dispatch(self, event):
        """Call each event subscriber with the event.
        """
        if self.event_subscribers is None:
            return
        for event_subscriber in self.event_subscribers:
            event_subscriber(event)
        return

    def connectionLost(self, reason=connectionDone):
        """Called when the connection is shut down.

        Clear any circular references here, and any external references
        to this Protocol.  The connection has been closed.

        Clear left over remaining bytes because they won't be continued
        properly upon reconnection.

        @type reason: L{twisted.python.failure.Failure}
        """
        self.status = DISCONNECTED
        self.remaining_bytes = b''
        if reason.check(error.ConnectionLost):
            self.dispatch(actions.ConnectionLost(self, reason=reason))
        else:  
            self.dispatch(actions.Disconnected(self, reason=reason))

    def dataReceived(self, data):
        """
        Parse the NATS.io protocol from chunks of data streaming from
        the connected gnatsd.

        The server settings will be set and connect will be sent with this
        client's info upon an INFO, which should happen when the
        transport connects.

        Registered message callback functions will be called with MSGs
        once parsed.

        PONG will be called upon a ping.

        An exception will be raised upon an ERR from gnatsd.

        An +OK doesn't do anything.
        """
        if self.remaining_bytes:
            data = self.remaining_bytes + data
            self.remaining_bytes = b''

        data_buf = BufferedReader(BytesIO(data))
        while True:
            command = data_buf.read(4)
            if command == b"-ERR":
                raise NatsError(data_buf.read())
            elif command == b"+OK\r":
                val = data_buf.read(1)
                if val != b'\n':
                    self.remaining_bytes += command
                    break
            elif command == b"MSG ":
                val = data_buf.readline()
                if not val:
                    self.remaining_bytes += command
                    break
                if not val.endswith(b'\r\n'):
                    self.remaining_bytes += command + val
                    break

                meta_data = val.split(b" ")
                n_bytes = int(meta_data[-1])
                subject = meta_data[0].decode()
                if len(meta_data) == 4:
                    reply_to = meta_data[2].decode()
                elif len(meta_data) == 3:
                    reply_to = None
                else:
                    self.remaining_bytes += command + val
                    break

                sid = meta_data[1].decode()

                if sid in self.sids:
                    on_msg = self.sids[sid]
                else:
                    on_msg = self.on_msg

                payload = data_buf.read(n_bytes)
                if len(payload) != n_bytes:
                    self.remaining_bytes += command + val + payload
                    break

                if on_msg:
                    on_msg(nats_protocol=self, sid=sid, subject=subject,
                           reply_to=reply_to, payload=payload)
                else:
                    stdout.write(command.decode())
                    stdout.write(val.decode())
                    stdout.write(payload.decode())

                self.dispatch(
                    actions.ReceivedMsg(
                        sid, self, 
                        subject=subject, 
                        payload=payload, 
                        reply_to=reply_to)
                    )

                if sid in self.unsubs:
                    self.unsubs[sid] -= 1
                    if self.unsubs[sid] <= 0:
                        self._forget_subscription(sid)
                        self.dispatch(actions.UnsubMaxReached(sid, protocol=self))

                payload_post = data_buf.readline()
                if payload_post != b'\r\n':
                    self.remaining_bytes += (command + val + payload
                                             + payload_post)
                    break
            elif command == b"PING":
                self.dispatch(actions.ReceivedPing(self))
                self.pong()
                val = data_buf.readline()
                if val != b'\r\n':
                    self.remaining_bytes += command + val
                    break
            elif command == b"PONG":
                self.pout -= 1
                self.dispatch(actions.ReceivedPong(self, outstanding_pings=self.pout))
                val = data_buf.readline()
                if val != b'\r\n':
                    self.remaining_bytes += command + val
                    break
            elif command == b"INFO":
                val = data_buf.readline()
                if not val.endswith(b'\r\n'):
                    self.remaining_bytes += command + val
                    break
                settings = json.loads(val.decode('utf8'))
                self.server_settings = ServerInfo(**settings)
                self.dispatch(actions.ReceivedInfo(
                    self, server_info=self.server_settings))
                self.status = CONNECTED
                self.pout = 0
                self.sids = {}
                self.connect()
                if self.on_connect_d:
                    self.on_connect_d.callback(self)
                    self.on_connect_d = None
            else:
                self.dispatch(actions.UnhandledCommand(self, command=command))
                val = data_buf.read()
                self.remaining_bytes += command + val
            if not data_buf.peek(1):
                break

    def connect(self):
        """
        Tell the NATS server about this client and it's options.
        """
        action = actions.SendConnect(self, client_info=self.client_info)
        payload = 'CONNECT {}\r\n'.format(json.dumps(
            self.client_info.asdict_for_connect()
            , separators=(',', ':')))

        self.transport.write(payload.encode())
        self.dispatch(action)

    def pub(self, subject,  payload, reply_to=None):
        """
        Publish a payload of bytes to a subject.

        @param subject: The destination subject to publish to.
        @param reply_to: The reply inbox subject that subscribers can use
         to send a response back to the publisher/requestor.
        @param payload: The message payload data, in bytes.
        """
        action = actions.SendPub(self, subject,  payload, reply_to)
        reply_part = ""
        if reply_to:
            reply_part = "{} ".format(reply_to)

        # TODO: deal with the payload if it is bigger than the server max.
        op = "PUB {} {}{}\r\n".format(
            subject, reply_part, len(payload)).encode()
        op += payload + b'\r\n'
        self.transport.write(op)
        self.dispatch(action)

    def apply_subscriptions(self):
        """
        Subscribe all the subscriptions and unsubscribe all of 
        the unsubscriptions.

        Builds the state of subscriptions and unsubscriptions 
        with max messages.
        """
        if self.status == CONNECTED:
            for sid, sub_args in self.subscriptions.items():
                self.sub(
                    sub_args.subject, sub_args.sid, 
                    sub_args.queue_group, sub_args.on_msg)
                if sid in self.unsubs:
                    self.unsub(sid, max_msgs=self.unsubs[sid])

    def sub(self, subject, sid, queue_group=None, on_msg=None):
        """
        Subscribe to a subject.

        @param subject: The subject name to subscribe to.
        @param sid: A unique alphanumeric subscription ID.
        @param queue_group: If specified, the subscriber will
         join this queue group.
         @param on_msg: A callable that takes the following params:
             @param nats_protocol: An instance of NatsProtocol.
             @param sid: A unique alphanumeric subscription ID.
             @param subject: A valid NATS subject.
             @param reply_to: The reply to.
             @param payload: Bytes of the payload.
        """
        self.sids[sid] = on_msg
        self.subscriptions[sid] = SubscriptionArgs(
            subject, sid, queue_group, on_msg)
        self.dispatch(actions.SendSub(
            sid=sid,
            protocol=self,
            subject=subject,
            queue_group=queue_group,
            on_msg=on_msg
        ))

        queue_group_part = ""
        if queue_group:
            queue_group_part = "{} ".format(queue_group)

        op = "SUB {} {}{}\r\n".format(subject, queue_group_part, sid)
        self.transport.write(op.encode('utf8'))

    def _forget_subscription(self, sid):
        """Undeclare a subscription. Any on_msg declared for the subscription 
        will no longer be called.
        If a apply_subscriptions is called,
        which it is during a reconnect, These subscriptions will not be 
        established. """
        if sid in self.unsubs:
            del self.unsubs[sid]
        if sid in self.subscriptions:
            del self.subscriptions[sid]
        if sid in self.sids:
            del self.sids[sid]

    def unsub(self, sid, max_msgs=None):
        """
        Unsubcribes the connection from the specified subject,
        or auto-unsubscribes after the specified
        number of messages has been received.

        @param sid: The unique alphanumeric subscription ID of
         the subject to unsubscribe from.
        @type sid: str
        @param max_msgs: Optional number of messages to wait for before
         automatically unsubscribing.
        @type max_msgs: int
        """
        action = actions.SendUnsub(sid=sid, protocol=self, max_msgs=max_msgs)
        max_msgs_part = ""
        if max_msgs:
            max_msgs_part = "{}".format(max_msgs)
            self.unsubs[sid] = max_msgs
        else:
            self._forget_subscription(sid)

        op = "UNSUB {} {}\r\n".format(sid, max_msgs_part)
        self.transport.write(op.encode('utf8'))
        self.dispatch(action)

    def ping(self):
        """
        Send ping.
        """
        op = b"PING\r\n"
        self.transport.write(op)
        self.pout += 1
        self.dispatch(actions.SendPing(self, outstanding_pings=self.pout))

    def pong(self):
        """
        Send pong.
        """
        op = b"PONG\r\n"
        self.transport.write(op)
        self.dispatch(actions.SendPong(self))

    def request(self, sid, subject):
        """
        Make a synchronous request for a subject.

        Make a reply to.
        Subscribe to the subject.
        Make a Deferred and add it to the inbox under the reply to.
        Do auto unsubscribe for one message.
        """
        raise NotImplementedError()



from tests.base import BaseTest


class TestPub(BaseTest):
    def test_pub(self):
        """
        Ensure pub sends the protocol PUB message with the given subject
        and payload.
        """
        def msg_handler(*args, **kwargs):
            pass
        self.nats_protocol.pub("aQueue", b"Do something!")
        self.assertEqual(self.transport.getvalue(),
                         b"PUB aQueue 13\r\nDo something!\r\n")

    def test_pub_with_reply_to(self):
        """
        Ensure pub sends the protocol PUB message with the reply_to
        subject included.
        """
        def msg_handler(*args, **kwargs):
            pass
        self.nats_protocol.pub("aQueue", b"Do something!", 'inbox')
        self.assertEqual(self.transport.getvalue(),
                         b"PUB aQueue inbox 13\r\nDo something!\r\n")




from twisted.trial import unittest

import txnats
from txnats import actions

class IsNatsProtocolTest(unittest.TestCase):
    maxDiff = None
    def test_received_ping(self):
        """
        Ensure ReceivedPing can be instanciated.
        """
        actions.ReceivedPing(protocol=txnats.io.NatsProtocol())

    def test_received_pong(self):
        """
        Ensure ReceivedPong takes the protocol.
        """
        actions.ReceivedPong(txnats.io.NatsProtocol(), outstanding_pings=0)    

    def test_send_ping(self):
        """
        Ensure SendPing takes the protocol.
        """
        actions.SendPing(txnats.io.NatsProtocol(), outstanding_pings=1)

    def test_sub_removed(self):
        actions.UnsubMaxReached("4", txnats.io.NatsProtocol())

    def test_send_unsub_immediate(self):
        "Ensure SendUnsub accepts no max_msgs"
        actions.SendUnsub("5", txnats.io.NatsProtocol())

    def test_send_unsub(self):
        actions.SendUnsub("5", txnats.io.NatsProtocol(), 3)
    
    def test_send_sub(self):
        actions.SendSub("6", txnats.io.NatsProtocol(), "a_subject")

    def test_send_pub(self):
        actions.SendPub(txnats.io.NatsProtocol(), "a_subject", b"hello", "callme")


from . import _meta


__version__ = _meta.version
__version_info__ = _meta.version_info

from . import protocol as io
from .protocol import NatsProtocol
from .backoff import Backoff

__all__ = [
    'io',
    'NatsProtocol',
    'Backoff'
]



import json
from io import BytesIO

from twisted.trial import unittest
from twisted.internet import task
from twisted.internet import defer

from twisted.python import failure
from twisted.internet import error

import txnats
from txnats import actions

from tests.base import BaseTest

class TestPartitionTolerance(BaseTest):

    def test_reconnect(self):
        """
        Ensure reconnect is called upon connection lost.
        """
        def msg_handler(*args, **kwargs):
            pass

        self.call_count = 0

        def reconnect(nats_protocol):
            self.call_count += 1
            self.assertIsInstance(nats_protocol, txnats.io.NatsProtocol)

        def event_subscriber(event):
            if isinstance(event, actions.ConnectionLost):
                reconnect(event.protocol)

        nats_protocol = txnats.io.NatsProtocol(
            own_reactor=self.reactor, 
            event_subscribers=[event_subscriber])
        nats_protocol.transport = BytesIO()

        self.nats_protocol.status = txnats.io.CONNECTED
        self.nats_protocol.sub('inbox', "1", queue_group="aQueueGroup",
                               on_msg=msg_handler)
        connectionLostFailure = failure.Failure(error.ConnectionLost())
        nats_protocol.connectionLost(reason=connectionLostFailure)
        self.reactor.advance(2)
        self.assertEqual(nats_protocol.status, txnats.io.DISCONNECTED)
        self.assertEqual(self.call_count, 1)


class TestDataReceived(BaseTest):
    def test_queue_group_subscribe(self):
        """
        Ensure a given queue group is put in between the subject and
        subscription id in the SUB command.
        """
        def msg_handler(*args, **kwargs):
            pass
        self.nats_protocol.sub('inbox', "1",
                               queue_group="aQueueGroup",
                               on_msg=msg_handler)
        self.assertEqual(self.transport.getvalue(),
                         b"SUB inbox aQueueGroup 1\r\n")

    def test_split_msg_payload(self):
        """
        Ensure a MSG command split across dataReceived within the payload is
        still processed the same, even if the payload ends with newline bytes.
        """
        self.msg_handler_called = False

        def msg_handler(nats_protocol=None, sid=None,
                        subject=None, reply_to=None, payload=None):
            self.msg_handler_called = True
            self.assertEqual(nats_protocol, self.nats_protocol)
            self.assertEqual(sid, "1")
            self.assertEqual(subject, "mysubject")
            self.assertEqual(reply_to, "inbox1")
            self.assertEqual(payload, b'h\r\nello')

        self.nats_protocol.sub('inbox', "1", on_msg=msg_handler)
        self.assertEqual(self.transport.getvalue(), b"SUB inbox 1\r\n")
        self.nats_protocol.dataReceived(
            b"MSG mysubject 1 inbox1 7\r\nh\r\n"
        )
        self.nats_protocol.dataReceived(
            b"ello\r\n"
        )
        self.assertTrue(self.msg_handler_called)

    def test_split_msg_with_partial_fake_msg_in_payload(self):
        """
        Ensure a MSG command split across dataReceived within the payload is
        still processed the same, even if the payload ends with newline bytes
        contains bytes that match the beginning of another command.
        """
        self.msg_handler_called = False

        def msg_handler(nats_protocol=None, sid=None,
                        subject=None, reply_to=None, payload=None):
            self.msg_handler_called = True
            self.assertEqual(nats_protocol, self.nats_protocol)
            self.assertEqual(sid, "1")
            self.assertEqual(subject, "mysubject")
            self.assertEqual(reply_to, "inbox1")
            self.assertEqual(
                payload, b"h\r\nello\r\nMSG asubject 3 breply 6\r\nsausage")

        self.nats_protocol.sub('inbox', "1", on_msg=msg_handler)
        self.assertEqual(self.transport.getvalue(), b"SUB inbox 1\r\n")
        self.nats_protocol.dataReceived(
            b"MSG mysubject 1 inbox1 41\r\nh\r\n"
        )
        self.nats_protocol.dataReceived(
            b"ello\r\nMSG asubject 3 breply 6\r\nsausage\r\n"
        )
        self.assertTrue(self.msg_handler_called)

    def test_split_msg_command(self):
        """
        Ensure a MSG command split within the first few bytes is handled.
        """
        self.msg_handler_called = False

        def msg_handler(nats_protocol=None, sid=None,
                        subject=None, reply_to=None, payload=None):
            self.msg_handler_called = True
            self.assertEqual(nats_protocol, self.nats_protocol)
            self.assertEqual(sid, "1")
            self.assertEqual(subject, "mysubject")
            self.assertEqual(reply_to, "inbox1")
            self.assertEqual(payload, b'h\r\nello')

        self.nats_protocol.sub('inbox', "1", on_msg=msg_handler)
        self.assertEqual(self.transport.getvalue(), b"SUB inbox 1\r\n")
        self.nats_protocol.dataReceived(
            b"MS"
        )
        self.nats_protocol.dataReceived(
            b"G mysubject 1 inbox1 7\r\nh\r\nello\r\n"
        )
        self.assertTrue(self.msg_handler_called)

    def test_split_stream(self):
        """
        Ensure a command split across multiple dataReceived calls is parsed
        and handled the same as commands wholely within one dataReceived.

        Ensure the next data received after one in where there was a previous
        protocol split is processed normally.
        """
        self.nats_protocol.dataReceived(
            b"PI"
        )
        self.nats_protocol.dataReceived(
            b"NG\r\n"
        )
        self.assertEqual(self.transport.getvalue(), b"PONG\r\n")
        self.nats_protocol.dataReceived(
            b"PING\r\n"
        )
        self.assertEqual(self.transport.getvalue(), b"PONG\r\nPONG\r\n")

    @defer.inlineCallbacks
    def test_info(self):
        """
        Ensure upon receiving an INFO operation, the server info is
        parsed and saved and a CONNECT operation is sent.

        """
        info_data = "INFO {}\r\n".format(
            json.dumps(
                {u'auth_required': False,
                    u'go': u'go1.5.2',
                    u'host': u'0.0.0.0',
                    u'max_payload': 1048576,
                    u'port': 4222,
                    u'server_id': u'16dd1049f122d8d3d148894074423d48',
                    u'ssl_required': False,
                    u'tls_required': False,
                    u'tls_verify': False,
                    u'version': u'0.7.2'}
            )
        )
        info_data = info_data.encode()
        self.nats_protocol.dataReceived(info_data)
        command = self.transport.getvalue()[:8]
        self.assertEqual(command, b'CONNECT ')
        client_info = json.loads(self.transport.getvalue()[8:].decode())
        self.assertEqual(
            client_info,
            {
                "verbose": True,
                "pedantic": False,
                "ssl_required": False,
                "auth_token": None,
                "user": "",
                "pass": "",
                "name": "txnats",
                "lang": "py.twisted",
                "version": txnats.__version__,
            }
        )

        self.assertEqual(
            self.nats_protocol.server_settings,
            txnats.config_state.ServerInfo(
                auth_required=False,
                go=u'go1.5.2',
                host=u'0.0.0.0',
                max_payload=1048576,
                port=4222,
                server_id=u'16dd1049f122d8d3d148894074423d48',
                ssl_required=False,
                tls_required=False,
                tls_verify=False,
                version=u'0.7.2'
            ))
        yield



from io import BytesIO

from twisted.trial import unittest
from twisted.internet import task
from twisted.internet import defer


import txnats


class BaseTest(unittest.TestCase):
    maxDiff = None

    @defer.inlineCallbacks
    def setUp(self):
        """
        Create protocol.

        Add any fixtures.

        :return:
        """
        self.reactor = task.Clock()
        self.nats_protocol = txnats.io.NatsProtocol(
            own_reactor=self.reactor)
        self.nats_protocol.transport = BytesIO()
        self.transport = self.nats_protocol.transport
        yield

    def tearDown(self):
        """
        Any tear down.

        :return:
        """




import attr
from twisted.trial import unittest

import txnats
from txnats.validators import is_instance_of_nats_protocol
from txnats.validators import is_subject
from txnats.validators import is_subject_id


@attr.s
class Foo(object):
    subject = attr.ib(validator=is_subject)
    sid = attr.ib(validator=is_subject_id)
    protocol = attr.ib(default=None, 
        validator=attr.validators.optional(
            is_instance_of_nats_protocol
        )
    )



class IsNatsProtocolTest(unittest.TestCase):
    maxDiff = None
    def test_foo_valid(self):
        """
        Ensure Foo.protocol may be a NatsProtocol.
        """
        Foo("A.*", "33", protocol=txnats.io.NatsProtocol())

    def test_wildcard_delimited_valid(self):
        Foo("A.*", "33", protocol=txnats.io.NatsProtocol())
        Foo("A.*.thing", "33", protocol=txnats.io.NatsProtocol())
        Foo("A.>", "33", protocol=txnats.io.NatsProtocol())
        Foo("A.*", "33", protocol=txnats.io.NatsProtocol())

    def test_wildcard_invalid(self):
        with self.assertRaises(ValueError):
            Foo("A.>ewr", "33", protocol=txnats.io.NatsProtocol())

        with self.assertRaises(ValueError):
            Foo("A>", "33", protocol=txnats.io.NatsProtocol())
        
        with self.assertRaises(ValueError):
            Foo("A*.la", "33", protocol=txnats.io.NatsProtocol())
        
        with self.assertRaises(ValueError):
            Foo("A.*la", "33", protocol=txnats.io.NatsProtocol())
    
    def test_foo_none(self):
        """
        Ensure Foo.protocol may be None.
        """
        Foo("B.2.>", "33")

    def test_foo_invalid(self):
        """
        Ensure any other value is invalid for Foo.protocol.
        """
        with self.assertRaises(ValueError):
            Foo("C3", "33", protocol=4)

    def test_foo_subject_not_alnum(self):
        """
        Ensure a subject not composed of strictly alphanumeric chars is
        invalid.
        """
        with self.assertRaises(ValueError):
            Foo("C 3", "33")

    def test_foo_subject_not_empty(self):
        """
        Ensure an empty string subject is not valid.
        """
        with self.assertRaises(ValueError):
            Foo("", "33")




from __future__ import absolute_import

import attr

from .validators import is_subject
from .validators import is_subject_id


@attr.s(slots=True)
class SubscriptionArgs(object):
    subject = attr.ib(validator=is_subject)
    sid = attr.ib(validator=is_subject_id)
    queue_group = attr.ib(
        default=None, 
        validator=attr.validators.optional(is_subject))
    on_msg = attr.ib(default=None)


@attr.s(slots=True, frozen=True)
class ServerInfo(object):
    server_id = attr.ib(default=None)
    version = attr.ib(default=None)
    go = attr.ib(default=None)
    host = attr.ib(default=None)
    port = attr.ib(default=None)
    auth_required = attr.ib(default=None)
    ssl_required = attr.ib(default=None)
    tls_required = attr.ib(default=None)
    tls_verify = attr.ib(default=None)
    max_payload = attr.ib(default=None)


@attr.s(slots=True, frozen=True)
class ClientInfo(object):
    verbose = attr.ib(validator=attr.validators.instance_of(bool))
    pedantic = attr.ib(validator=attr.validators.instance_of(bool))
    ssl_required = attr.ib(validator=attr.validators.instance_of(bool))
    auth_token = attr.ib(default=None)
    user = attr.ib(default=None)
    password = attr.ib(default=None)
    name = attr.ib(default="txnats")
    lang = attr.ib(default="py.twisted")
    version = attr.ib(default="")

    def asdict_for_connect(self):
        client_info = attr.asdict(self, 
            filter=attr.filters.exclude(
                attr.fields(self.__class__).password))
        client_info["pass"] = self.password
        return client_info



from __future__ import absolute_import

from twisted.internet import defer
from twisted.internet.endpoints import connectProtocol
from twisted.internet import error
from twisted.logger import Logger
log = Logger()

from .protocol import NatsProtocol
from . import actions


def sleep(own_reactor, seconds):
    d = defer.Deferred()
    own_reactor.callLater(seconds, d.callback, seconds)
    return d


@defer.inlineCallbacks
def connect(point, protocol, backoff, max_retries=100):
    while backoff.retries <= max_retries:
        log.debug("tries {}".format(backoff.retries))
        log.debug("connecting..")
        try:
            yield connectProtocol(point, protocol)
            backoff.reset_delay()
            log.debug(".connected")
            defer.returnValue(None)
        except (error.ConnectError, error.DNSLookupError):
            delay = backoff.get_delay()
            log.debug("connection failed, sleep for {}".format(delay))
            log.error()
            yield sleep(protocol.reactor, delay)


def make_reconnector(point, backoff, max_retries=100):
    """
    Makes and returns an event subscriber function that, in the event of
    a disconnect, it calls this module's connect function, and upon
    sending the client connect, apply the subscription state.
    """
    def reconnector(event):
        if isinstance(event, actions.SendConnect):
            event.protocol.apply_subscriptions()
        elif isinstance(event, actions.ConnectionLost):
            if event.protocol.ping_loop.running:
                event.protocol.ping_loop.stop()
            if not event.reason.check(error.ConnectionDone):
                protocol = NatsProtocol(
                    own_reactor=event.protocol.reactor,
                    verbose=event.protocol.verbose, 
                    pedantic=event.protocol.client_info.pedantic,
                    ssl_required=event.protocol.client_info.ssl_required,
                    auth_token=event.protocol.client_info.auth_token,
                    user=event.protocol.client_info.user,
                    password=event.protocol.client_info.password,
                    on_msg=event.protocol.on_msg,
                    #on_connect=event.protocol.on_connect,
                    event_subscribers=event.protocol.event_subscribers, 
                    subscriptions=event.protocol.subscriptions,
                    unsubs=event.protocol.unsubs)
                connect(point, protocol, backoff, max_retries)
    return reconnector


def ping_loop_controller(event):
    """
    Upon connect, it starts the ping loop and upon disconnect,
    stops the ping loop.
    """
    if isinstance(event, actions.SendConnect):
        if not event.protocol.ping_loop.running:
            event.protocol.ping_loop.start(10, now=True)
    elif isinstance(event, (actions.Disconnected, actions.ConnectionLost)):
        if event.protocol.ping_loop.running:
            event.protocol.ping_loop.stop()



import os
import sys

try:
    from setuptools import find_packages
    from setuptools import setup
    packages = find_packages(exclude='tests')
except ImportError:
    from distutils.core import setup
    packages = ['txnats']

from pip.req import parse_requirements
try:
        from pip.download import PipSession
except ImportError as e:
        raise ImportError("cannot import name PipSession, "
                          "Please update pip to >= 1.5")

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


def read(fname):
    """
    Open and read a filename in this directory.
    :param fname: `str` file name in this directory

    Returns contents of file fname
    """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def get_version():
    import imp

    with open('txnats/_meta.py', 'rb') as fp:
        mod = imp.load_source('_meta', 'txnats', fp)

    return mod.version


def get_requirement_items(filename):
    """
    Get install requirements from a file.
    :param filename: A file path to a requirements file.
    :return: A list of requirement items.
    :rtype: A `list` of :class:`pip.req.InstallRequirement`
    """
    with PipSession() as session:
        return list(parse_requirements(filename, session=session))


def get_requirements(filename):
    """
    Get requirements from a file and
    convert to a list of strings.
    :param filename:
    :return:
    """
    reqs = get_requirement_items(filename)

    return [str(r.req) for r in reqs]


def get_install_requires():
    return get_requirements('requirements.txt')


def get_test_requires():
    return get_requirements('requirements_dev.txt')


try:
    license_info = open('LICENSE').read()
except:
    license_info = 'The MIT License (MIT)'

description = "NATS messaging system client using Twisted #microservices"
long_description = description
if os.path.exists('.generated_README.rst'):
    long_description = open('.generated_README.rst').read()

setup_args = dict(
    name="txnats",
    version=get_version(),
    author="John W Lockwood IV",
    author_email="john.lockwood@workiva.com",
    description=description,
    license=license_info,
    keywords="data",
    url="https://github.com/johnwlockwood/txnats",
    package_dir={'txnats': 'txnats'},
    packages=packages,
    install_requires=get_install_requires(),
    tests_require=get_test_requires(),
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)

if __name__ == '__main__':
    setup(**setup_args)



import re

SID_SUBJECT_TYPES = (type(u""), type(""))


def is_instance_of_nats_protocol(instance, attribute, value):
    from .protocol import NatsProtocol
    if not isinstance(value, NatsProtocol):
        raise ValueError(
            "The value of {} must be an instance of NatsProtocol!".format(
                attribute.name))


def is_subject_id(instance, attribute, value):
    if not isinstance(value, SID_SUBJECT_TYPES) or not value.isalnum():
        raise ValueError(
            "The value of {} must be an alpha numeric string! got: '{}''".format(
                attribute.name, value))


def is_subject(instance, attribute, value):
    if not isinstance(value, SID_SUBJECT_TYPES) or not re.match(
        "^([\w\d]*)(\.[\w\d]+)?(\.\*\.[\w\d]*)?(\.\*)?(^>)?(\.>)?$", value) or value == "":
        raise ValueError(
            "The value of {} must be an alpha numeric string with the "
            "exception of '*' '.' and '>'! got: {}".format(
                attribute.name, value))
